use clap::{arg, Command};
use gpauth::saml_auth;
use gpcommon::portal::{Portal, Prelogin, SamlPrelogin};
use serde_json::json;

fn cli() -> Command {
    Command::new("gpauth")
        .about("GlobalProtect-openconnect authentication helper")
        .arg_required_else_help(true)
        .arg(arg!(<SERVER> "The GlobalProtect server"))
}

#[tokio::main]
async fn main() -> anyhow::Result<()> {
    fern::Dispatch::new()
        .format(|out, message, record| {
            out.finish(format_args!(
                "[{} {} {}] {}",
                humantime::format_rfc3339_millis(std::time::SystemTime::now()),
                record.level(),
                record.target(),
                message
            ))
        })
        .level(log::LevelFilter::Info)
        .chain(std::io::stderr())
        .apply()?;

    let matches = cli().get_matches();
    let server = matches.get_one::<String>("SERVER").expect("Missing server");
    let address = if server.starts_with("https://") || server.starts_with("http://") {
        server.to_string()
    } else {
        format!("https://{}", server)
    };

    let portal = Portal::new(&address);
    let saml_request = match portal.prelogin().await? {
        Prelogin::Saml(SamlPrelogin { request, .. }) => request,
        _ => anyhow::bail!("Prelogin response is not SAML"),
    };
    let auth_data = saml_auth(&portal, Some(&saml_request)).await?;

    // Output the auth data as JSON, so that the client can parse it
    println!("{}", json!(auth_data));

    Ok(())
}
